Module: MilkTea::LSP::Server::ServerCompletion

Included in:
MilkTea::LSP::Server
Defined in:
lib/milk_tea/lsp/server/completion.rb

Constant Summary collapse

MAX_COMPLETION_ITEMS =
200
COMPLETION_KEYWORDS =
%w[
  let var const function async struct enum flags variant union
  type interface opaque if while for match return import public static
  extending external fn event attribute defer break continue pass
  unsafe consuming implements include link foreign proc
  editable in out inout as when inline await static_assert emit
].freeze
TYPE_CONSTRUCTOR_KEYWORDS =
%w[
  ptr ref span array dyn Option Result Task SoA str_buffer const_ptr simd
].freeze