Module: MilkTea::Lexer::Recovery
- Included in:
- MilkTea::Lexer
- Defined in:
- lib/milk_tea/core/lexer/recovery.rb
Overview
Error-recovery heuristics: detecting a top-level declaration line to resynchronize on after an unterminated grouping or heredoc.
Error-recovery heuristics: detecting a top-level declaration line to resynchronize on after an unterminated grouping or heredoc.