Module: MilkTea::Parse::Recovery
- Included in:
- MilkTea::Parser
- Defined in:
- lib/milk_tea/core/parser/recovery.rb
Constant Summary collapse
- TOP_LEVEL_RECOVERY_START_TYPES =
%i[ module import at public attribute const var type struct union enum flags variant interface opaque extending foreign async function event external static_assert link include compiler_flag ].freeze